/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier, SafeRiver/LaBRI

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

#include "solver.h"

using namespace Glucose;
/* ========================================================================== */
/*                             Solver functions                               */
/* ========================================================================== */
void init_solver(bool elim_ll, bool verbose) {
  solver = new SimpSolver();
  solver->parsing = 0;
  solver->verbosity = 0;
  solver->verbEveryConflicts = 10000;
  solver->showModel = true;
  solver->certifiedUNSAT = false;
  solver->use_elim = elim_ll;
  solver->verb_mc = verbose;

  assumptions.clear();
}

lbool check_sat(bool verb) {
  clock_t clk_diff, clk;
  
  if (verb)
    printf("[glucose] Conflicts : %8lu | Clauses : %8d | Learnts : %8d | "
	   "Unary : %8lu | Binary : %8lu\n",
	   solver->conflicts, solver->nClauses(), solver->nLearnts(),
	   solver->nbUn, solver->nbBin);
  
  clk = clock();
  lbool ret = solver->solveLimited(assumptions);
  clk_diff = clock() - clk;
  assumptions.clear();
  
  if (verb) {
    printf(ret == l_True ? "[glucose] SAT (%6.2f sec)\n" :
	                   (ret == l_False ? "[glucose] UNSAT (%6.2f sec)\n" :
			                     "[glucose] INDETERMINATE (%6.2f sec)\n"),
	   (float)(clk_diff)/(float)(CLOCKS_PER_SEC));
  }
  return ret;
}

void assume(int lit) {
  int var;
  var = abs(lit);
  if (var >= solver->nVars()) {
    printf("assume | unknown var\n");
    exit(-1);

  } else {
    assumptions.push( (lit > 0) ? mkLit(var) : ~mkLit(var) );
  }
}

void ucl(int a) {
  vec<Lit> lits;
  int var;
  var = abs(a);
  while (var >= solver->nVars()) solver->newVar();
  lits.push( (a > 0) ? mkLit(var) : ~mkLit(var) );

  assert(!solver->isEliminated(abs(a)));

  solver->addClause_(lits);
}

void bcl(int a, int b) {
  vec<Lit> lits;
  int var;
  var = abs(a);
  while (var >= solver->nVars()) solver->newVar();
  lits.push( (a > 0) ? mkLit(var) : ~mkLit(var) );

  var = abs(b);
  while (var >= solver->nVars()) solver->newVar();
  lits.push( (b > 0) ? mkLit(var) : ~mkLit(var) );
  assert(!solver->isEliminated(abs(a)));
  assert(!solver->isEliminated(abs(b)));

  solver->addClause_(lits);
}



void tcl(int a, int b, int c) {
  vec<Lit> lits;
  int var;
  var = abs(a);
  while (var >= solver->nVars()) solver->newVar();
  lits.push( (a > 0) ? mkLit(var) : ~mkLit(var) );

  var = abs(b);
  while (var >= solver->nVars()) solver->newVar();
  lits.push( (b > 0) ? mkLit(var) : ~mkLit(var) );

  var = abs(c);
  while (var >= solver->nVars()) solver->newVar();
  lits.push( (c > 0) ? mkLit(var) : ~mkLit(var) );

  assert(!solver->isEliminated(abs(a)));
  assert(!solver->isEliminated(abs(b)));
  assert(!solver->isEliminated(abs(c)));

  solver->addClause_(lits);
}

void cl_lhs2rhs(unsigned lhs, std::list<unsigned> &rhs) {
  assert(lhs > 1);
  
  std::list<unsigned>::iterator it_rhs;
  vec<Lit> lits;
  int var = lhs >> 1;

  while (var >= solver->nVars()) solver->newVar();
  aiger_sign(lhs) ? lits.push(mkLit(var)) : lits.push(~mkLit(var));
  assert(!solver->isEliminated(var));

  for (it_rhs = rhs.begin(); it_rhs != rhs.end(); ++it_rhs) {
    assert(*it_rhs > 1);
    var = *it_rhs >> 1;
    while (var >= solver->nVars()) solver->newVar();
    assert(var < solver->nVars());
    assert(!solver->isEliminated(var));
    lits.push( aiger_sign(*it_rhs) ? ~mkLit(var) : mkLit(var) );
  }

  solver->addClause_(lits);
}

void cl_rhs2lhs(unsigned lhs, std::list<unsigned> &rhs) {
  assert(lhs > 1);
  
  std::list<unsigned>::iterator it_rhs;
  vec<Lit> lits;
  int var = lhs >> 1;

  while (var >= solver->nVars()) solver->newVar();
  aiger_sign(lhs) ? lits.push(~mkLit(var)) : lits.push(mkLit(var));
  assert(!solver->isEliminated(var));

  for (it_rhs = rhs.begin(); it_rhs != rhs.end(); ++it_rhs) {
    assert(*it_rhs > 1);
    var = *it_rhs >> 1;
    while (var >= solver->nVars()) solver->newVar();
    assert(var < solver->nVars());
    assert(!solver->isEliminated(var));
    lits.push( aiger_sign(*it_rhs) ? mkLit(var) : ~mkLit(var) );
  }

  solver->addClause_(lits);
}


void bigAnd2Solver(unsigned lhs, prop_tr_un_fl *ptuf) {
  assert(ptuf->op == AND);
  vec<Lit> lits_tmp;
  vec<Lit> lits;

  int var_tmp;
  int var = lhs >> 1;
  while (var >= solver->nVars()) solver->newVar();
  assert(!solver->isEliminated(var));

  list_ptuf_e *it = ptuf->lp->head;
  
  while(it != NULL) {
    assert(it->ptuf->op == VAR);
    lits_tmp.clear();

    assert(it->ptuf->lit > 1);
    var_tmp = it->ptuf->lit >> 1;
    while (var_tmp >= solver->nVars()) solver->newVar();
    assert(var_tmp < solver->nVars());
    assert(!solver->isEliminated(var_tmp));
    lits_tmp.push( aiger_sign(it->ptuf->lit) ? ~mkLit(var_tmp) : mkLit(var_tmp));
    lits_tmp.push( aiger_sign(lhs) ? mkLit(var) : ~mkLit(var));
    solver->addClause_(lits_tmp);

    lits.push( aiger_sign(it->ptuf->lit) ? mkLit(var_tmp) : ~mkLit(var_tmp));

    it = it->next;
  }
  
  lits.push( aiger_sign(lhs) ? ~mkLit(var) : mkLit(var));
  solver->addClause_(lits);
}



std::pair<std::list<std::list<unsigned>>, std::list<std::list<unsigned>>> pt2cl(prop_tr_un_fl *ptuf) {
  std::list<unsigned> li;
  std::list<std::list<unsigned>> lli_f;
  std::list<std::list<unsigned>> lli_s;
  
  if (ptuf->op == VAR) {
    li.push_back(ptuf->lit);
    lli_f.push_back(li);
    lli_s.push_back(li);

  } else if (ptuf->lp->size == 1) {
    assert(0);
  } else {
    assert(ptuf->lp->size > 1);
    std::list<std::pair<std::list<std::list<unsigned>>, std::list<std::list<unsigned>>>> lli_pl;
    
    list_ptuf_e *it = ptuf->lp->head;
    
    while(it != NULL) {
      std::pair<std::list<std::list<unsigned>>, std::list<std::list<unsigned>>> lli_p = pt2cl(it->ptuf);
      lli_pl.push_back(lli_p);
      
      it = it->next;
    }
    
    std::list<std::pair<std::list<std::list<unsigned>>, std::list<std::list<unsigned>>>>::iterator it_llip;
    std::list<std::list<unsigned>>::iterator it_lli;
    std::list<std::list<unsigned>>::iterator it_lli2;
    std::list<unsigned>::iterator it_li;

    if (ptuf->op == AND) {
      
      for( it_llip = lli_pl.begin(); it_llip != lli_pl.end(); it_llip++) {
	for (it_lli = it_llip->first.begin(); it_lli != it_llip->first.end(); it_lli++) {
	  lli_f.push_back(*it_lli);
	}
      }
      
      std::list<std::list<unsigned>> lli_op1_s = lli_pl.front().second;
      lli_pl.pop_front();
      while (!lli_pl.empty()) {
	std::list<std::list<unsigned>> lli_op2_s = lli_pl.front().second;
	lli_pl.pop_front();
	for (it_lli = lli_op1_s.begin(); it_lli != lli_op1_s.end(); it_lli++) {
	  for (it_lli2 = lli_op2_s.begin(); it_lli2 != lli_op2_s.end(); it_lli2++) {
	    li.clear();
	    
	    for (it_li = it_lli->begin(); it_li != it_lli->end(); it_li++) {
	      li.push_back(*it_li);
	    }
	    for (it_li = it_lli2->begin(); it_li != it_lli2->end(); it_li++) {
	      li.push_back(*it_li);
	    }
	    lli_s.push_back(li);
	  }
	}
	lli_op1_s = lli_s;
	lli_s.clear();
      }
      lli_s = lli_op1_s;
      
    } else {

      for( it_llip = lli_pl.begin(); it_llip != lli_pl.end(); it_llip++) {
	for (it_lli = it_llip->second.begin(); it_lli != it_llip->second.end(); it_lli++) {
	  lli_s.push_back(*it_lli);
	}

      }
      
      std::list<std::list<unsigned>> lli_op1_f = lli_pl.front().first;
      lli_pl.pop_front();
      while (!lli_pl.empty()) {
	std::list<std::list<unsigned>> lli_op2_f = lli_pl.front().first;
	lli_pl.pop_front();
	for (it_lli = lli_op1_f.begin(); it_lli != lli_op1_f.end(); it_lli++) {
	  for (it_lli2 = lli_op2_f.begin(); it_lli2 != lli_op2_f.end(); it_lli2++) {
	    li.clear();
	    
	    for (it_li = it_lli->begin(); it_li != it_lli->end(); it_li++) {
	      li.push_back(*it_li);
	    }
	    for (it_li = it_lli2->begin(); it_li != it_lli2->end(); it_li++) {
	      li.push_back(*it_li);
	    }
	    lli_f.push_back(li);
	  }
	}
	lli_op1_f = lli_f;
	lli_f.clear();
      }
      lli_f = lli_op1_f;
      
    }
  }
  return std::pair<std::list<std::list<unsigned>>, std::list<std::list<unsigned>>>(lli_f, lli_s);
}


void ands2clauses(int time) {
  ands_def* tmp_def;
  dll_elem<ands_def> *tmp_e;

  // Ands def
  tmp_e = list_ands[time]->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    
    assert(tmp_def->lhs > 1);

    if (tmp_def->rhs_tr_fl->op == AND && tmp_def->rhs_tr_fl->lp->tail->ptuf->op == VAR) {
      // Optimized big and def to clauses
      bigAnd2Solver(tmp_def->lhs, tmp_def->rhs_tr_fl);
      
     } else {
      std::pair<std::list<std::list<unsigned>>, std::list<std::list<unsigned>>> lli_p =
	pt2cl(tmp_def->rhs_tr_fl);

      std::list<std::list<unsigned>>::iterator it_lli;

      for (it_lli = lli_p.first.begin(); it_lli != lli_p.first.end(); it_lli++) {
	cl_lhs2rhs(tmp_def->lhs, *it_lli);
      }

      for (it_lli = lli_p.second.begin(); it_lli != lli_p.second.end(); it_lli++) {
	cl_rhs2lhs(tmp_def->lhs, *it_lli);
      }
    }
    
    tmp_e = tmp_e->next;
  }

}
